"""Provide exceptions for aiovlc."""

from __future__ import annotations


class AIOVLCError(Exception):
    """Represent a common error for aiovlc."""


class ConnectError(AIOVLCError):
    """Represent a connection error for aiovlc."""


class ConnectReadError(ConnectError):
    """The client failed to read."""

    def __init__(self, error: Exception, partial_bytes: bytes | None = None) -> None:
        """Set up error."""
        message = f"Failed to read: {error}."

        if partial_bytes is not None:
            message = f"{message} Partial bytes read: {partial_bytes!r}"

        super().__init__(message)
        self.partial_bytes = partial_bytes


class AuthError(AIOVLCError):
    """Represent an authentication error."""


class CommandError(AIOVLCError):
    """Represent a command error."""


class CommandParameterError(CommandError):
    """Represent an error with a parameter when calling the command."""


class CommandParseError(CommandError):
    """Represent an error when parsing the command output."""



"""Provide commands for aiovlc."""

from __future__ import annotations

import contextlib
from dataclasses import dataclass, field
import re
from typing import TYPE_CHECKING, Generic, Literal, TypeVar

from ..const import LOGGER
from ..exceptions import (
    AuthError,
    CommandError,
    CommandParameterError,
    CommandParseError,
)

if TYPE_CHECKING:
    from ..client import Client

T = TypeVar("T")
DEFAULT_COMMAND_READ_TERMINATOR = "> "


@dataclass
class Command(Generic[T]):
    """Represent a VLC command."""

    log_command: bool = field(init=False, default=True)
    prefix: str = field(init=False)
    read_terminator: str = field(init=False, default=DEFAULT_COMMAND_READ_TERMINATOR)

    async def send(self, client: Client) -> T:
        """Send the command."""
        return await self._send_command(client)

    async def _send_command(self, client: Client) -> T:
        """Send the command."""
        command_string = self.build_command()

        async with client.command_lock:
            if self.log_command:
                LOGGER.debug("Sending command: %s", command_string.strip())
            await client.write(command_string)
            raw_output = await client.read(self.read_terminator)

        command_output = raw_output.split("\r\n")[:-1]
        LOGGER.debug("Command output: %s", command_output)

        if command_output:
            if re.match(
                r"Unknown command `.*'\. Type `help' for help\.",
                command_output[0],
            ):
                raise CommandError("Unknown Command")
            if command_error := re.match(r"Error in.*", command_output[0]):
                raise CommandError(command_error.group())

        return self.parse_output(command_output)

    def build_command(self) -> str:
        """Return the full command string."""
        return f"{self.prefix}\n"

    def parse_output(self, output: list[str]) -> T:  # noqa: ARG002
        """Parse command output."""
        # Disable mypy to have cleaner code in child classes.
        return None  # type: ignore[return-value]


@dataclass
class CommandOutput:
    """Represent a command output."""


@dataclass
class Add(Command[None]):
    """Represent the add command."""

    prefix = "add"
    playlist_item: str

    def build_command(self) -> str:
        """Return the full command string."""
        return f"{self.prefix} {self.playlist_item}\n"


@dataclass
class Clear(Command[None]):
    """Represent the clear command."""

    prefix = "clear"


@dataclass
class Enqueue(Command[None]):
    """Represent the enqueue command."""

    prefix = "enqueue"
    playlist_item: str

    def build_command(self) -> str:
        """Return the full command string."""
        return f"{self.prefix} {self.playlist_item}\n"


@dataclass
class GetLengthOutput(CommandOutput):
    """Represent the get length command output."""

    length: int


@dataclass
class GetLength(Command[GetLengthOutput]):
    """Represent the get length command."""

    prefix = "get_length"

    def parse_output(self, output: list[str]) -> GetLengthOutput:
        """Parse command output."""
        try:
            if not (length_string := output[0]):
                return GetLengthOutput(length=0)
            length = int(length_string)
        except (IndexError, ValueError) as err:
            raise CommandParseError("Could not get length.") from err
        return GetLengthOutput(length=length)


@dataclass
class GetTimeOutput(CommandOutput):
    """Represent the get time command output."""

    time: int


@dataclass
class GetTime(Command[GetTimeOutput]):
    """Represent the get time command."""

    prefix = "get_time"

    def parse_output(self, output: list[str]) -> GetTimeOutput:
        """Parse command output."""
        try:
            if not (time_string := output[0]):
                return GetTimeOutput(time=0)
            time = int(time_string)
        except (IndexError, ValueError) as err:
            raise CommandParseError("Could not get time.") from err
        return GetTimeOutput(time=time)


@dataclass
class InfoOutput(CommandOutput):
    """Represent the info command output."""

    data: dict[str | int, dict[str, str | int | float]] = field(default_factory=dict)


@dataclass
class Info(Command[InfoOutput]):
    """Represent the info command."""

    prefix = "info"

    def parse_output(self, output: list[str]) -> InfoOutput:
        """Parse command output."""
        data: dict[str | int, dict[str, str | int | float]] = {}
        section: int | str = "unknown"
        for line in output:
            if line[0] == "+":
                # Example: "+----[ Stream 5 ]" or "+----[ Meta data ]"  # noqa: ERA001
                if "end of stream info" in line:
                    continue
                section = line.split("[")[1].replace("]", "").strip().split(" ")[1]
                with contextlib.suppress(ValueError):
                    section = int(section)
                data[section] = {}
            elif line[0] == "|":
                # Example: "| Description: Closed captions 4"  # noqa: ERA001
                if len(line[2:]) == 0:
                    continue
                value: int | float | str = "unknown"
                key, value = line[2:].split(":", 1)
                try:
                    value = int(value)
                except ValueError:
                    try:
                        value = float(value)
                    except ValueError:
                        value = value.strip()  # type: ignore[union-attr]
                data[section][key.strip()] = value
            else:
                raise CommandParseError(f"Unexpected line in info output: {line}")
        return InfoOutput(data=data)


@dataclass
class PasswordOutput(CommandOutput):
    """Represent the password command output."""

    response: str


@dataclass
class Password(Command[PasswordOutput]):
    """Represent the password command."""

    log_command = False
    prefix = ""
    password: str
    read_terminator = "\n"

    def build_command(self) -> str:
        """Return the full command string."""
        return f"{self.prefix}{self.password}\n"

    def parse_output(self, output: list[str]) -> PasswordOutput:
        """Parse command output."""
        response: str = ""
        for line in output:
            if not line:
                continue
            response = line
            parsed_output = response.lower()
            if "wrong password" in parsed_output:
                raise AuthError("Failed to login to VLC.")
            if "welcome" not in parsed_output:
                raise CommandError(f"Unexpected password response: {response}")

        if not response:
            raise CommandError("Empty password response")
        return PasswordOutput(response)

    async def send(self, client: Client) -> PasswordOutput:
        """Send the command."""
        await client.read("Password: ")
        password_output = await super().send(client)
        if DEFAULT_COMMAND_READ_TERMINATOR in password_output.response:
            password_output.response = password_output.response.strip(
                DEFAULT_COMMAND_READ_TERMINATOR,
            )
            return password_output
        # Read until prompt
        await client.read(DEFAULT_COMMAND_READ_TERMINATOR)
        return password_output


@dataclass
class Next(Command[None]):
    """Represent the next command."""

    prefix = "next"


@dataclass
class Pause(Command[None]):
    """Represent the pause command."""

    prefix = "pause"


@dataclass
class Play(Command[None]):
    """Represent the play command."""

    prefix = "play"


@dataclass
class Prev(Command[None]):
    """Represent the prev command."""

    prefix = "prev"


@dataclass
class Random(Command[None]):
    """Represent the random command."""

    prefix = "random"
    mode: Literal["on", "off"] | None = None
    VALID_MODES = (None, "on", "off")

    def build_command(self) -> str:
        """Return the full command string."""
        if self.mode not in self.VALID_MODES:
            raise CommandParameterError(f"Parameter mode not in {self.VALID_MODES}")
        mode = "" if self.mode is None else f" {self.mode}"
        return f"{self.prefix}{mode}\n"


@dataclass
class Seek(Command[None]):
    """Represent the seek command."""

    prefix = "seek"
    seconds: int

    def build_command(self) -> str:
        """Return the full command string."""
        return f"{self.prefix} {self.seconds}\n"


@dataclass
class StatusOutput(CommandOutput):
    """Represent the status command output."""

    audio_volume: int
    state: str
    input_loc: str | None = None


@dataclass
class Status(Command[StatusOutput]):
    """Represent the status command."""

    prefix = "status"

    def parse_output(self, output: list[str]) -> StatusOutput:
        """Parse command output."""
        input_loc: str | None = None
        if len(output) == 3:
            input_loc_item = output.pop(0)
            input_loc = "%20".join(input_loc_item.split(" ")[3:-1])
        if len(output) == 2:
            audio_volume = int(float(output[0].split(" ")[3].replace(",", ".")))
            state = output[1].split(" ")[2]
        else:
            raise CommandParseError("Could not get status.")
        return StatusOutput(audio_volume=audio_volume, state=state, input_loc=input_loc)


@dataclass
class Stop(Command[None]):
    """Represent the stop command."""

    prefix = "stop"


@dataclass
class VolumeOutput(CommandOutput):
    """Represent the volume command output."""

    audio_volume: int


@dataclass
class Volume(Command[VolumeOutput]):
    """Represent the volume command."""

    prefix = "volume"

    def parse_output(self, output: list[str]) -> VolumeOutput:
        """Parse command output."""
        try:
            audio_volume = int(float(output[0].replace(",", ".")))
        except (IndexError, ValueError) as err:
            raise CommandParseError("Could not get volume.") from err
        return VolumeOutput(audio_volume=audio_volume)


@dataclass
class SetVolume(Command[None]):
    """Represent the set volume command."""

    prefix = "volume"
    volume: int
    VALID_VOLUME = range(500)

    def build_command(self) -> str:
        """Return the full command string."""
        try:
            volume = int(self.volume)
        except ValueError as err:
            raise CommandParameterError(
                f"Invalid volume parameter: {self.volume}",
            ) from err
        if volume not in self.VALID_VOLUME:
            raise CommandParameterError(f"Parameter volume not in {self.VALID_VOLUME}")
        return f"{self.prefix} {volume}\n"



"""CLI for the VLC client."""

import asyncio
from collections.abc import Awaitable, Callable
from typing import Annotated

from rich import print
import typer

from .client import Client
from .exceptions import AIOVLCError
from .model.command import Info, Status

app = typer.Typer()

ClientFactory = Callable[[], Awaitable[Client]]


@app.command()
def client(
    password: Annotated[
        str,
        typer.Option(help="Password of the connection.", prompt=True, hide_input=True),
    ],
    host: Annotated[str, typer.Option(help="Host of the connection.")] = "localhost",
    port: Annotated[int, typer.Option(help="Port of the connection.")] = 4212,
) -> None:
    """Start a client."""

    async def client_factory() -> Client:
        """Return a client."""
        return Client(password, host=host, port=port)

    run_client(client_factory)


def run_client(client_factory: ClientFactory) -> None:
    """Run a client."""
    print("Starting client")
    try:
        asyncio.run(start_client(client_factory))
    except KeyboardInterrupt:
        pass
    finally:
        print("Exiting CLI")


async def start_client(client_factory: ClientFactory) -> None:
    """Start the client."""
    vlc_client = await client_factory()

    try:
        async with vlc_client:
            await vlc_client.login()
            await handle_client(vlc_client)
    except AIOVLCError as err:
        print("Error:", err)


async def handle_client(vlc_client: Client) -> None:
    """Handle the client calls."""
    while True:
        command = Status()
        output = await command.send(vlc_client)
        print("Received:", output)
        info_command = Info()
        info_output = await info_command.send(vlc_client)
        print("Received:", info_output)
        await asyncio.sleep(10)



"""Provide a client for aiovlc."""

from __future__ import annotations

import asyncio
from types import TracebackType
from typing import Literal, Self

from .const import LOGGER
from .exceptions import ConnectError, ConnectReadError
from .model.command import (
    Add,
    Clear,
    Enqueue,
    GetLength,
    GetLengthOutput,
    GetTime,
    GetTimeOutput,
    Info,
    InfoOutput,
    Next,
    Password,
    PasswordOutput,
    Pause,
    Play,
    Prev,
    Random,
    Seek,
    SetVolume,
    Status,
    StatusOutput,
    Stop,
    Volume,
    VolumeOutput,
)

IAC = bytes([255])  # "Interpret As Command"
TERMINATOR = "\n"


class Client:
    """Represent a client for aiovlc."""

    # pylint: disable=too-many-public-methods

    def __init__(
        self,
        password: str,
        host: str = "localhost",
        port: int = 4212,
    ) -> None:
        """Set up the client client."""
        self.host = host
        self.password = password
        self.port = port
        self.command_lock = asyncio.Lock()
        self._reader: asyncio.StreamReader | None = None
        self._writer: asyncio.StreamWriter | None = None

    async def __aenter__(self) -> Self:
        """Connect the client with context manager."""
        await self.connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Disconnect the client with context manager."""
        await self.disconnect()

    async def connect(self) -> None:
        """Connect the client."""
        try:
            self._reader, self._writer = await asyncio.open_connection(
                host=self.host,
                port=self.port,
            )
        except OSError as err:
            raise ConnectError(f"Failed to connect: {err}") from err

    async def disconnect(self) -> None:
        """Disconnect the client."""
        if self._writer is None:
            raise RuntimeError("Client is not connected")
        try:
            self._writer.close()
            await self._writer.wait_closed()
        except OSError:
            pass

    async def read(self, read_until: str = TERMINATOR) -> str:
        """Return a decoded message."""
        if self._reader is None:
            raise RuntimeError("Client is not connected")

        try:
            read = await self._reader.readuntil(read_until.encode("utf-8"))
        except asyncio.LimitOverrunError as err:
            raise ConnectReadError(err) from err
        except asyncio.IncompleteReadError as err:
            raise ConnectReadError(err, err.partial) from err
        except OSError as err:
            raise ConnectError(f"Failed to read: {err}") from err

        LOGGER.debug("Bytes read: %s", read)

        # Drop IAC command and read again.
        if IAC in read:
            return await self.read(read_until)
        return read.decode("utf-8")

    async def write(self, command: str) -> None:
        """Write a command to the connection."""
        if self._writer is None:
            raise RuntimeError("Client is not connected")

        try:
            self._writer.write(command.encode("utf-8"))
            await self._writer.drain()
        except OSError as err:
            raise ConnectError(f"Failed to write: {err}") from err

    async def login(self) -> PasswordOutput:
        """Login."""
        return await Password(self.password).send(self)

    async def add(self, playlist_item: str) -> None:
        """Send the add command."""
        await Add(playlist_item).send(self)

    async def clear(self) -> None:
        """Send the clear command."""
        await Clear().send(self)

    async def enqueue(self, playlist_item: str) -> None:
        """Send the enqueue command."""
        await Enqueue(playlist_item).send(self)

    async def get_length(self) -> GetLengthOutput:
        """Send the get_length command."""
        return await GetLength().send(self)

    async def get_time(self) -> GetTimeOutput:
        """Send the get_time command."""
        return await GetTime().send(self)

    async def info(self) -> InfoOutput:
        """Send the info command."""
        return await Info().send(self)

    async def next(self) -> None:
        """Send the next command."""
        await Next().send(self)

    async def pause(self) -> None:
        """Send the pause command."""
        await Pause().send(self)

    async def play(self) -> None:
        """Send the play command."""
        await Play().send(self)

    async def prev(self) -> None:
        """Send the prev command."""
        await Prev().send(self)

    async def random(self, mode: Literal["on", "off"] | None = None) -> None:
        """Send the random command."""
        await Random(mode).send(self)

    async def seek(self, seconds: int) -> None:
        """Send the seek command."""
        await Seek(seconds).send(self)

    async def set_volume(self, volume: int) -> None:
        """Send the volume command with a parameter to set volume."""
        await SetVolume(volume).send(self)

    async def status(self) -> StatusOutput:
        """Send the status command."""
        return await Status().send(self)

    async def stop(self) -> None:
        """Send the stop command."""
        await Stop().send(self)

    async def volume(self) -> VolumeOutput:
        """Send the volume command."""
        return await Volume().send(self)
